local _G = _G or GLOBAL
local S = _G.STRINGS.MODS.MYHOME
local T = _G.TUNING.MYHOME
local space = string.rep(" ", 9)

-- 須存檔的參數
local SAVEPARAMS = {
    "friends",
    "players",
    "names",
    "config",
}

local myhome = Class(function(self, inst)

--Public
self.inst = inst

-- params
self.friends = {}   -- [MID] = {[friendsMID] = 1, ...}
self.names = {}     -- [MID] = name
self.config = {}    -- [sum] = int

-- Tag list
-- lock     轉換鎖
-- animlock 動畫鎖
-- [userid] 轉換使用者
-- [a]f[b]  a是b朋友
-- mh_mid[..mid]

-- 判斷兩人是否為朋友關係(Client只會判斷牆壁是否為朋友的)
function self:GetFriendsByMid(t_mid, r_mid)
    return t_mid ~= nil and self.friends[t_mid] ~= nil and r_mid ~= nil and self.friends[t_mid][r_mid] ~= nil
end

-- 透過Mid取得玩家名稱(Client只會用來顯示焦點牆壁的建造人)
function self:GetPlayerNameByMid(mid)
    local name = S.A19
    if mid and self.names[mid] then
        name = self.names[mid]
    end
    return name
end

inst.net_playerssum = net_ushortint(inst.GUID, "playerssum", "playerssum_dirty")
if T.DISPLAY_NAME ~= 0 then inst.net_names = net_string(inst.GUID, "names", "names_dirty") end
inst.net_friends = net_string(inst.GUID, "friends", "friends_dirty")

if TheNet:GetIsClient() then
----- ↓↓↓↓↓ Only Clent ↓↓↓↓↓ -----

    inst:ListenForEvent("playerssum_dirty", function(inst)
        self.config.playerssum = inst.net_playerssum:value()
        if self.config.selfmid == 0 then
            self.config.selfmid = nil
        end
    end)
    if T.DISPLAY_NAME ~= 0 then
        inst:ListenForEvent("names_dirty", function(inst)
            local names = {}
            string.gsub(inst.net_names:value(), "([^\\]+)", function(w)
                table.insert(names, w)
            end)
            self.names = names
        end)
    end
    inst:ListenForEvent("friends_dirty", function(inst)
        local list = {}
        for text in string.gmatch(inst.net_friends:value(), '[^|]+') do
            local key = 0
            for n in string.gmatch(text, '[^:]+') do
                if key == 0 then
                    key = tonumber(n)
                    list[key] = {}
                else
                    list[key][tonumber(n)] = 1
                end
            end
        end
        self.friends = list
    end)

    -- 查詢玩家的MID(Client只會查詢自己的MID，所以改成遍歷自己的Tag)
    function self:GetMidByUserid(userid)
        local mid = self.config.selfmid
        if mid == nil and ThePlayer ~= nil then

            for i = 1, self.config.playerssum or 0 do
                if ThePlayer:HasTag("mh_mid"..i) then
                    mid = i
                    break
                end
            end
            if mid == nil then
                mid = 0
            end
            self.config.selfmid = mid
        end
        return mid > 0 and mid or nil
    end

    ----- Client找牆壁OID的方法 -----
    function self:GetWallOidByTag(wall)
        FindEntity(wall, 3, function(guy)
            if guy ~= nil and guy.components.wall ~= nil and guy.components.wall.inst.oid ~= nil then
                local i = guy.components.wall.inst.oid
                if wall:HasTag("mh_oid"..i) == true then
                    wall.oid = i
                    return
                end
            end
        end, {"wall"})
        if wall.oid ~= nil then return end
        for i = 0, self.config.playerssum or 0 do
            if wall:HasTag("mh_oid"..i) == true then
                if i == 0 then
                    i = nil
                end
                wall.oid = i
                return
            end
        end
    end

    return
end
----- ↓↓↓↓↓ Only Server ↓↓↓↓↓ -----

-- params
self.players = {}   -- [psn(userid的剪版)] = MID

self.handle_config = {
    player = nil,   -- 申請人
    mid = nil,      -- 申請人mid
    total = nil,    -- 總牆數
    count = {},     -- 自己、朋友、別人
    timeout = nil,  -- 超時function
    -- 紀錄x,z最小/大值
    x1 = nil,
    z1 = nil,
    x2 = nil,
    z2 = nil,
}

-- 重置轉換配置
local function cancel()
    for _, v in ipairs(TheSim:FindEntities(0, 0, 0, 1e4, {"mh_focus"})) do
        v:RemoveTag("mh_focus")
    end
    for _, v in ipairs(TheSim:FindEntities(0, 0, 0, 1e4, {"mh_guide"})) do
        v:RemoveTag("mh_guide")
    end
    for _, v in ipairs(TheSim:FindEntities(0, 0, 0, 1e4, {"mh_first"})) do
        v:RemoveTag("mh_first")
    end

    inst:RemoveTag("animlock")
    inst:DoTaskInTime(5, function ()
        inst:RemoveTag(self.handle_config.player.userid)
        inst:RemoveTag("lock")
    end)
end

-- 逾時處理
local function timeout()
    self:Talker(self.handle_config.player, S.A1)
    cancel()
end

-- MyHome 轉換成功
local function success(pos)
    -- 第幾個家
    if self.config.sum == nil then
        self.config.sum = 0
    end
    local newhid = self.config.sum + 1
    self.config.sum = newhid

    -- 轉換處理
    inst:StartThread(function()
        for n, v in ipairs(TheSim:FindEntities(0, 0, 0, 1e4, {"mh_focus"})) do
            local target = v.components.wall
            if v:HasTag("mh_first") then
                target.inst:SetHid(newhid)
                target.inst.pos = pos
            end
            target.inst:AddLock(newhid)
            if target.inst.oid == nil then
                target.inst:SetOid(self.handle_config.mid)
            end

            if n % 50 == 0 then
                Sleep(.5)
            end
        end

        self:Talker(self.handle_config.player, S.A2, 3)
        if T.ANNOUNCE then
            TheNet:Announce(string.format(S.A3, self.handle_config.player:GetDisplayName(), self.handle_config.total))
        end
        cancel()
    end)
end

-- MyHome處理器
function self:HandleMyHome(target, doer)
    local c = self.handle_config
    -- 不可用: 未執行、指向沒有指引的牆、操作者非申請人
    if inst:HasTag("animlock") ~= true or target:HasTag("mh_guide") ~= true or inst:HasTag(doer.userid) ~= true then return end

    -- 取消逾時、紀錄當前牆
    if c.timeout ~= nil then
        c.timeout:Cancel()
        c.timeout = nil
    end

    -- 消除指引、選中當前牆
    local x, y, z = target.Transform:GetWorldPosition()
    if x == nil or z == nil then
        cancel()
    end
    local total = c.total + 1
    c.total = total
    for _, v in ipairs(TheSim:FindEntities(x, y, z, 3.5, {"mh_guide"})) do
        v:RemoveTag("mh_guide")
    end
    target:AddTag("mh_focus")

    -- 抓出x, z最大最小值
    if x > c.x2 then
        c.x2 = x
    elseif x < c.x1 then
        c.x1 = x
    end
    if z > c.z2 then
        c.z2 = z
    elseif z < c.z1 then
        c.z1 = z
    end

    -- 判斷牆的所有者並說話
    if target.oid == nil or c.mid == target.oid then
        c.count[1] = c.count[1] + 1
    elseif self:GetFriendsByMid(target.oid, c.mid) == true then
        c.count[2] = c.count[2] + 1
    else
        c.count[3] = c.count[3] + 1
    end
    self:Talker(doer, string.format(S.A4, total, c.count[1], c.count[2], c.count[3]), 10)

    -- 選中後隱藏
    inst:DoTaskInTime(0.6, function ()
        target.components.wall.inst:SetNeonColour(2, 1)
    end)

    -- 判斷是否與頭相接
    if total >= T.MIN_WALL or T.ADMIN == true and doer.Network:IsServerAdmin() and total > 7 then
        if #TheSim:FindEntities(x, y, z, 1.5, {"mh_first"}) > 0 then
            -- 禁涵蓋判斷
            local exerr = {}
            local cx = (c.x1 + c.x2) / 2
            local cz = (c.z1 + c.z2) / 2
            local cr = math.max(c.x2 - cx, c.z2 - cz)
            local pos = {x=cx, z=cz, r=cr+3, x1=c.x1, z1=c.z1, x2=c.x2, z2=c.z2}

            if T.ADMIN == true and doer.Network:IsServerAdmin() then
                self:Talker(doer, S.A5, 10)
                return success(pos)
            else
                if T.RESTRICTED_AREA then
                    for _, v in ipairs(TheSim:FindEntities(cx, 0, cz, cr, {"mh_r1"})) do
                        if self:GetReverseMyhome(v, c.x1, c.z1, c.x2, c.z2, 0, {"mh_focus"}) then
                            table.insert(exerr, string.format(S.A6, tostring(v.name)))
                        end
                    end
                    for _, v in ipairs(TheSim:FindEntities(cx, 0, cz, cr + 30, {"mh_r2"})) do
                        if self:GetReverseMyhome(v, c.x1, c.z1, c.x2, c.z2, 30, {"mh_focus"}) then
                            table.insert(exerr, string.format(S.A7, tostring(v.name)))
                        end
                    end
                    if total < 60 then
                        for _, v in ipairs(TheSim:FindEntities(cx, 0, cz, cr, {"mh_r3"})) do
                            if self:GetReverseMyhome(v, c.x1, c.z1, c.x2, c.z2, 0, {"mh_focus"}) then
                                table.insert(exerr, string.format(S.A6, tostring(v.name)))
                            end
                        end
                    end
                end
                for _, v in ipairs(TheSim:FindEntities(cx, 0, cz, cr + 5, {"mh_lock"})) do
                    if self:GetReverseMyhome(v, c.x1, c.z1, c.x2, c.z2, 5, {"mh_focus"}) then
                        table.insert(exerr, S.A8)
                        break
                    end
                end

                -- 所有者比例判斷
                if #exerr == 0 then
                    if total * .2 < c.count[3] then
                        self:Talker(doer, S.A9, 10)
                    elseif total * .6 < c.count[2] then
                        self:Talker(doer, S.A10, 10)
                    else
                        self:Talker(doer, S.A5, 10)
                        return success(pos)
                    end
                else
                    self:Talker(doer, table.concat(exerr, "\n"), 10)
                end
            end
            return inst:DoTaskInTime(.5, cancel)
        end
    elseif total > T.MAX_WALL then
        self:Talker(doer, string.format(S.A11, T.MAX_WALL))
        return inst:DoTaskInTime(.5, cancel)
    end

    -- 添加指引
    local find = TheSim:FindEntities(x, y, z, 1.5, {"wall"}, {"mh_focus", "mh_lock"})
    local new_find = {}

    for _, v in ipairs(find) do
        if v.components.wall ~= nil then    -- 防止抓到不是牆卻有wall標籤的prefabs
            v:AddTag("mh_guide")
            v.components.wall.inst:SetNeonText(2)    -- Client端有自己的做法
            v.components.wall.inst:SetNeonColour(1, 2)
            table.insert(new_find, v)
        end
    end

    if #new_find == 0 then
        self:Talker(doer, string.format(S.A12, T.MIN_WALL))
        return inst:DoTaskInTime(3, function ()
            cancel()
        end)
    end

    if #new_find == 2 then -- 轉角自動選取處理
        for k, v in ipairs(new_find) do
            local tx, ty, tz = v.Transform:GetWorldPosition()
            local corner = TheSim:FindEntities(tx, ty, tz, 1.5, {"wall"}, {"mh_focus", "mh_lock"})
            if #corner == 2 then
                for _, q in ipairs(corner) do
                    if q ~= v and q:HasTag("mh_guide") then
                        table.remove(new_find, k % 2 + 1)
                    end
                end
            end
        end
    end

    if #new_find == 1 then -- 自動選取
        self:HandleMyHome(new_find[1].components.wall.inst, doer)
    else -- 添加逾時
        c.timeout = inst:DoTaskInTime(15, timeout)
    end
end

-- 申請MyHome
function self:ReceiverMyHome(target, doer)
    local mid = self:GetMidByUserid(doer.userid)
    -- 不可以申請: 沒有申請人、申請人錯誤、沒有目標、目標不是申請人造的
    if doer == nil or doer.components == nil or target == nil or target.oid ~= mid then
        return self:Talker(doer, S.A14)
    end
    local c = self.handle_config

    if inst:HasTag("lock") then
        if c.mid ~= mid then
            self:Talker(doer, string.format(S.A15, c.player:GetDisplayName()))
        end
        return
    end

    inst:AddTag("lock")
    inst:AddTag("animlock")
    inst:AddTag(doer.userid)

    target:AddTag("mh_guide")
    target:AddTag("mh_first")

    c.player, c.mid, c.total, c.count = doer, mid, 0, {0, 0, 0}

    local x, y, z = target.Transform:GetWorldPosition()
    if x == nil or z == nil then
        cancel()
    end
    c.x1, c.z1, c.x2, c.z2 = x, z, x, z

    self:Talker(doer, S.A16, 4)

    inst:DoTaskInTime(3, function ()
        self:HandleMyHome(target, doer)
    end)
end

-- 取消MyHome
function self:CancelMyHome(target, doer)
    -- 不可以申請: 目標非起始點、不是自己的
    if target == nil or target.hid == nil or (target.oid ~= self:GetMidByUserid(doer.userid) and T.ADMIN == true and not doer.Network:IsServerAdmin()) then
        return self:Talker(doer, S.A14)
    end

    -- 解除所有該MyHome
    for _, v in ipairs(TheSim:FindEntities(0, 0, 0, 1e4, {"mh_"..target.hid})) do
        v.components.wall.inst:RemoveLock()
        v.components.wall.inst:SetHid(nil)
    end

    self:Talker(doer, S.A17)
    if T.ANNOUNCE then
        TheNet:Announce(string.format(S.A18, doer:GetDisplayName()))
    end
end

-- 人物說話的安全檢查
function self:Talker(player, msg, sec)
    if player and player.components and player.components.talker then
        local talker = player.components.talker
        talker:Say(msg, sec or 3)
        return talker
    end
    return nil
end

-- 判斷是否在MyHome內或周圍
function self:GetReverseMyhome(traget, x1, z1, x2, z2, cr, tag)
    if traget == nil or traget.Transform == nil then return false end
    local rx, _, rz = traget.Transform:GetWorldPosition()
    if rx == nil or rz == nil then return false end
    if rx > x2 or rx < x1 or rz > z2 or rz < z1 then
        if cr == 0 or #TheSim:FindEntities(rx, 0, rz, cr, tag) > 0 then
            return false
        end
    end
    return true
end

-- 將兩人的userid轉為MID
function self:GetFriends(talker, recipient)
    return self:GetFriendsByMid(self:GetMidByUserid(talker, true), self:GetMidByUserid(recipient, true))
end

-- 將userid轉為MID並得到玩家名字
function self:GetPlayerName(userid)
    return self:GetPlayerNameByMid(self:GetMidByUserid(userid))
end

-- 查詢玩家的MID
function self:GetMidByUserid(userid, req)
    local mid = nil
    if userid ~= nil and type(userid)=="string" then
        local psn = string.sub(userid..space, 4,11)
        if self.players[psn] ~= nil then
            mid = self.players[psn]
        end
    end
    if req == true and mid == nil then
        mid = self:SetPlayerName(userid)
    end
    return mid
end

-- 透過userid取得player
function self:GetPlayerById(userid)
    for _, v in ipairs(AllPlayers) do
        if v ~= nil and v.userid and v.userid == userid then
            return v
        end
    end
    return nil
end

-- 授權別人時判斷遊戲編號對應的玩家
function self:GetPlayerBySid(sid)
    local isdedicated = not TheNet:GetServerIsClientHosted()
    local index = 1
    sid = tonumber(sid)

    for _, v in ipairs(TheNet:GetClientTable()) do
        if not isdedicated or v.performance == nil then
            if index == sid then
                return self:GetPlayerById(v.userid)
            end
            index = index + 1
        end
    end
    return nil
end

-- 請人授權時須要知道自己的遊戲編號
function self:GetSidByPlayer(player)
    local isdedicated = not TheNet:GetServerIsClientHosted()
    local index = 1

    for _, v in ipairs(TheNet:GetClientTable()) do
        if not isdedicated or v.performance == nil then
            if v.userid == player.userid then
                break
            end
            index = index + 1
        end
    end
    return index
end

-- 將兩人的userid轉為MID，然後開關朋友關係(呼叫這個方法表示雙方都必須有MID)
function self:SetFriends(talker, recipient, allow)
    local t_mid = self:GetMidByUserid(talker, true)
    local r_mid = self:GetMidByUserid(recipient, true)
    if self.friends[t_mid] == nil then
        self.friends[t_mid] = {}
    end

    allow = allow ~= false and 1 or nil

    if self.friends[t_mid][r_mid] ~= allow then
        self.friends[t_mid][r_mid] = allow
        self:SendFriends()
    end
end

-- 將所有玩家的名稱及userid=MID保存到兩個Server變量中
function self:SetPlayerName(userid)
    if userid ~= nil then
        local player = self:GetPlayerById(userid)
        if player ~= nil then
            local player_mid = self:GetMidByUserid(userid)
            local name = player:GetDisplayName()
            if player_mid then
                if self.names[player_mid] ~= name then
                    self.names[player_mid] = name
                    self:SendNames()
                end
            else
                table.insert(self.names, name)
                player_mid = #self.names
                self:SetPlayerMid(userid, player_mid)
                self:SendPlayers()
                self:SendNames()
            end
            player:AddTag("mh_mid"..player_mid) -- 為了Client
            return player_mid
        end
    end
    return nil
end

-- 新增userid到players烈表
function self:SetPlayerMid(userid, num)
    if userid ~= nil and type(userid)=="string" then
        local psn = string.sub(userid..space, 4,11)
        self.players[psn] = num
    end
end

function self:SendFriends()
    local ret = ""
    for k1, v1 in pairs(self.friends) do
        ret = ret.."|"..k1
        for k2, v2 in pairs(v1) do
            ret = ret..":"..k2
        end
    end
    ret = string.sub(ret, 2)
    inst.net_friends:set(ret)
end
function self:SendPlayers() -- 因為Client只會讀取自己的MID，所以採用Tag判斷法
    inst.net_playerssum:set(#self.names)
end
function self:SendNames()
    if T.DISPLAY_NAME ~= 0 then
        local ret = ""
        for _, v in ipairs(self.names) do
            v = string.gsub(v, "\\", "")
            local _, count = string.gsub(v, "[^\128-\193]", "")
            if count <= T.DISPLAY_NAME then
                ret = ret..v.."\\"
            else
                local n = T.DISPLAY_NAME
                for uchar in string.gfind(v, "[%z\1-\127\194-\244][\128-\191]*") do
                    ret = ret..uchar
                    n = n - 1
                    if n <= 0 then break end
                end
                ret = ret.."\\"
            end

        end
        if ret:len() > 7200 then
            T.DISPLAY_NAME = T.DISPLAY_NAME - 1
            print("[MOD|MYHOME]Names exceeds the maximum limit 7200 bytes, change to length "..tostring(T.DISPLAY_NAME))
            self:SendNames()
        else
            inst.net_names:set(ret)
        end
    end
end

function self:OnSave()
    self.config.ver = 134

    local save = {}
    for _, v in pairs(SAVEPARAMS) do
        save[v] = self[v]
    end
    return save
end

function self:OnLoad(data)
    for _, v in pairs(SAVEPARAMS) do
        if data[v] then
            for n, p in pairs(data[v]) do
                self[v][n] = p
            end
        end
    end
    self:SendPlayers()
    self:SendNames()
    self:SendFriends()
end

end)

----- Server收到Client端的焦點 -----
AddModRPCHandler("wall", "OnFocus", function(doer, x, y, z, sel)
    local ents = TheSim:FindEntities(x, y, z, .5, {"wall"})
    if #ents > 0 then
        for _, target in pairs(ents) do
            if sel then
                target.components.wall.inst:SendMyHomeFocus(doer)
            end
        end
    end
end)

return myhome
