local _G = _G or GLOBAL
local TheSim = _G.TheSim
local TheWorld = _G.TheWorld
local MH = _G.MH
local S = _G.STRINGS.MODS.MYHOME
local T = _G.TUNING.MYHOME
local NEONCOLOUR_MODE = {
    {   -- [1]導引下一個
        {.8, .8, .8, 1}, 3, nil, nil, .5
    },
    {   -- [3]已選取的標示
        {0, 0, 0, 0.1}, 0, nil, nil, .5
    },
}
local NEONTEXT_MODE = {
    {   -- [1]指標移入時顯示所有人
        {.5, 1, 0}, 30, 3, .5,
    },
    {   -- [2]導引下一個
        {1, 0, .5}, 52, 2, .5
    },
    {   -- [3]已選取的標示
        {.8, .5, .2}, 36, 2, .5
    },
}

----- 固定提示效果 -----
local animloop = {
    function(inst, cb)
        inst:StartThread(function() while MH.inst:HasTag("animlock") do Sleep(0.2) end cb() end)
    end,
    function(inst, cb)
        inst:StartThread(function() while inst:HasTag("mh_guide") do Sleep(0.2) end cb() end)
    end,
}
----- 改變顏色 -----
local function SetNeonColour(inst, mode, cb)
    if inst.lock_neoncolour ~= true then
        local anim = inst.AnimState
        if anim == nil then return nil end
        local mc1 = NEONCOLOUR_MODE[mode][1] or {1, 1, 1, 1}
        local light1 = NEONCOLOUR_MODE[mode][2] or 1
        local mc2 = NEONCOLOUR_MODE[mode][3] or {1, 1, 1, 1}
        local light2 = NEONCOLOUR_MODE[mode][4] or 0
        local sec = NEONCOLOUR_MODE[mode][5] or .5

        anim:SetMultColour(mc1[1],mc1[2],mc1[3],mc1[4])
        anim:SetLightOverride(light1)
        inst.lock_neoncolour = true

        if type(cb) == "number" then
            cb = animloop[cb]
        end
        local restore = function()
            anim:SetMultColour(mc2[1],mc2[2],mc2[3],mc2[4])
            anim:SetLightOverride(light2)
            inst.lock_neoncolour = nil
        end
        inst:DoTaskInTime(sec, function()
            if cb ~= nil then cb(inst, restore) else restore() end
        end)
    end
    return nil
end
----- 顯示文字 -----
local function SetNeonText(inst, text, mode, cb)
    if inst.lock_neontext ~= true and text ~= 0 then
        if type(text) == "number" then
            if text == 2 then
                text, mode, cb = S.A13, 2, 2
            end
        end

        local label
        if inst.Label then
            label = inst.Label
        elseif inst.entity then
            label = inst.entity:AddLabel()
        else return nil end
        local rgb = NEONTEXT_MODE[mode][1] or {1, 1, 1}
        local size = NEONTEXT_MODE[mode][2] or 24
        local offset = NEONTEXT_MODE[mode][3] or 2
        local sec = NEONTEXT_MODE[mode][4] or .5
        local text = text or ""
        label:SetFont(DEFAULTFONT)
        label:SetFontSize(size)
        label:SetWorldOffset(0, offset, 0)
        label:SetColour(rgb[1],rgb[2],rgb[3])
        label:SetText(text)
        label:Enable(true)
        inst.lock_neontext = true

        if type(cb) == "number" then
            cb = animloop[cb]
        end
        local restore = function()
            inst.Label:Enable(false)
            inst.lock_neontext = false
        end
        inst:DoTaskInTime(sec, function()
            if cb ~= nil then cb(inst, restore) else restore() end
        end)
    end
    return nil
end
----- 下面都是server端 -----
----- 發送焦點 -----
local function SendMyHomeFocus(inst, doer)
    _G.MH:HandleMyHome(inst, doer)
end

----- 添加鎖定 -----
local function AddLock(inst, lockid)
    inst.lock = tonumber(lockid)
    inst:AddTag("mh_lock")
    inst:AddTag("mh_"..inst.lock)
    inst:AddTag("notarget")

    if T.INVINCIBLE_TYPE ~= true and inst.hid == nil then      -- 如果不是true，只有起點可以無敵
        return true
    end

    local comp = inst.components
    if comp ~= nil then
        if comp.workable ~= nil and comp.workable.CanBeWorked ~= nil and comp.workable:CanBeWorked() == true then
            comp.workable:SetWorkable(false)
            comp.workable.Mh_WorkedBy = comp.workable.WorkedBy
            comp.workable.WorkedBy = function(worker, numworks) end
            comp.workable.Mh_workable = true
        end
        if comp.health ~= nil and comp.health.IsInvincible ~= nil and comp.health:IsInvincible() == false then
            comp.health:SetInvincible(true)
            comp.health.Mh_health = false
        end
        if comp.combat ~= nil and comp.combat.canattack ~= true then
            comp.combat.canattack = false
            comp.combat.Mh_canattack = true
        end
        if comp.burnable ~= nil then
            inst:RemoveTag("canlight")
            inst:AddTag("nolight")
            inst:AddTag("fireimmune")
            comp.burnable.extinguishimmediately = true
            if comp.burnable.flammability ~= nil then
                comp.burnable.Mh_flammability = comp.burnable.flammability
                comp.burnable.flammability = 0
            end
            if comp.propagator ~= nil and comp.propagator.flashpoint ~= nil then
                comp.propagator.Mh_flashpoint = comp.propagator.flashpoint
                comp.propagator.flashpoint = 0
            end
        end
    end
end

----- 取消鎖定 -----
local function RemoveLock(inst)
    inst:RemoveTag("mh_lock")
    inst:RemoveTag("mh_"..inst.lock)
    inst:RemoveTag("notarget")
    inst.lock = nil

    local comp = inst.components
    if comp.workable.Mh_workable ~= nil and comp.workable.SetWorkable ~= nil then
        comp.workable:SetWorkable(true)
        comp.workable.Mh_workable = nil
        comp.workable.WorkedBy = comp.workable.Mh_WorkedBy
        comp.workable.Mh_WorkedBy = nil
    end
    if comp.health.Mh_health ~= nil and comp.health.SetInvincible ~= nil then
        comp.health:SetInvincible(false)
        comp.health.Mh_health = nil
    end
    if comp.combat.Mh_canattack ~= nil then
        comp.combat.canattack = true
        comp.combat.Mh_canattack = nil
    end
    if inst.components.burnable ~= nil then
        inst:AddTag("canlight")
        inst:RemoveTag("nolight")
        inst:RemoveTag("fireimmune")
        if comp.burnable.Mh_flammability ~= nil then
            inst.components.burnable.flammability = comp.burnable.Mh_flammability
            comp.burnable.Mh_flammability = nil
        end
        if comp.propagator.Mh_flashpoint ~= nil then
            inst.components.propagator.flashpoint = comp.propagator.Mh_flashpoint
            comp.propagator.Mh_flashpoint = nil
        end
    end
end

----- 驅逐 -----
local function getout(inst)
    local p = inst.pos
    for _, v in ipairs(TheSim:FindEntities(p.x, 0, p.z, p.r, {"player"})) do
        local mid = MH:GetMidByUserid(v.userid)
        if mid ~= inst.oid and MH:GetFriendsByMid(inst.oid, mid) ~= true and not (T.ADMIN == true and v.Network:IsServerAdmin() == true) and MH:GetReverseMyhome(v, p.x1, p.z1, p.x2, p.z2, 0, {"mh_lock"}) then
            MH:Talker(v, string.format(S.B3, MH:GetPlayerNameByMid(inst.oid), MH:GetSidByPlayer(v)))

            local x, y, z = v.Transform:GetWorldPosition()
            local nx, nz = x, z
            local n = 0
            while n < 10 do
                local rand = math.random(5, math.max(p.r, 5))
                if math.abs(p.x - nx) < math.abs(p.z - nz) then
                    nx = nx + rand * (nx >= p.x and 1 or -1)
                else
                    nz = nz + rand * (nz >= p.z and 1 or -1)
                end
                if TheWorld.Map:IsAboveGroundAtPoint(nx, y, nz) then
                    inst:DoTaskInTime(0, getout)
                    break
                end
                n = n + 1
            end
            if n < 10 then
                v.Transform:SetPosition(nx, y, nz)
            end
        end
    end
end

----- 權限保存與加載 -----
local function SetOid(inst, newoid)
    local oldoid = inst.oid or 0
    inst:RemoveTag("mh_oid"..oldoid)
    inst.oid = newoid
    newoid = newoid or 0
    inst:AddTag("mh_oid"..newoid)
end

local function SetHid(inst, newhid)
    inst.hid = newhid
    if inst.canthere_task ~= nil then
        inst.canthere_task:Cancel()
        inst.canthere_task = nil
    end
    if inst.hid ~= nil then
        inst:AddTag("mh_ctrl")
        inst:DoTaskInTime(1.5, function()
            local p = inst.pos
            if inst.hid ~= nil and T.INTRUDERS == true and p.x~=nil and p.z~=nil and p.r~=nil and p.x1~=nil and p.z1~=nil and p.x2~=nil and p.z2~=nil then
                -- 驅趕入侵者
                inst.canthere_task = inst:DoPeriodicTask(4, getout)
            end
        end)
    else
        inst:RemoveTag("mh_ctrl")
    end
end

local function OnSave(inst,data,seal)
    if seal~=true and type(inst.QueueSave)=='table' then
        for _,save in ipairs(inst.QueueSave) do
            if save ~= nil then save(inst,data,true) end
        end
    end
    for _, v in ipairs({
        "oid",  -- 建造者的MID
        "lock", -- 是否鎖定
        "hid",  -- 是否起點
        "pos",  -- 起點紀錄中點
    }) do
        if inst[v] ~= nil then
            data[v] = inst[v]
        end
    end
end

local function OnLoad(inst,data,seal)
    if seal~=true and type(inst.QueueLoad)=='table' then
        for _,load in ipairs(inst.QueueLoad) do
            if load ~= nil then load(inst,data,true) end
        end
    end
    if data ~= nil then
        SetOid(inst, data.oid)
        if data.lock ~= nil then
            AddLock(inst, data.lock)
        end
        if data.hid ~= nil then
            SetHid(inst, data.hid)
            if data.pos ~= nil and data.pos.x1 ~= nil then
                inst.pos = data.pos
            else
                -- 替沒有中點的起點加上中點(相容v1.1.1以下的MyHome)
                -- 替沒有最大最小x,z的點加上(相容v1.3.6以下的MyHome)
                inst:DoTaskInTime(1, function()
                    if inst.lock == nil then return end
                    local x, y, z = inst.Transform:GetWorldPosition()
                    local x1, z1, x2, z2 = x, z, x, z
                    for _, v in ipairs(TheSim:FindEntities(0, 0, 0, 1e4, {"mh_"..inst.lock})) do
                        local vx, vy, vz = v.Transform:GetWorldPosition()
                        if vx > x2 then
                            x2 = vx
                        elseif vx < x1 then
                            x1 = vx
                        end
                        if vz > z2 then
                            z2 = vz
                        elseif vz < z1 then
                            z1 = vz
                        end
                    end
                    local cx = (x1 + x2) / 2
                    local cz = (z1 + z2) / 2
                    local cr = math.max(x2 - cx, z2 - cz)
                    inst.pos = {x=cx, z=cz, r=cr+3, x1=x1, z1=z1, x2=x2, z2=z2}
                end)
            end
        end
    end
end

-- Tag list
-- mh_focus 轉換模式下的選中
-- mh_guide 轉換模式下的導引
-- mh_first 轉換模式下的起點
-- mh_oid[?] 所有者mid
-- mh_ctrl   起點牆
-- mh_lock   MyHome模式
-- mh_[?]    家的編號
return  Class(function(self, inst)
self.inst = inst
inst.SetNeonColour = SetNeonColour
inst.SetNeonText = SetNeonText

if not TheWorld.ismastersim then
    return
end
inst.SendMyHomeFocus = SendMyHomeFocus

inst.AddLock = AddLock
inst.RemoveLock = RemoveLock

inst.SetOid = SetOid
inst.SetHid = SetHid


if type(inst.QueueSave) ~= 'table' then
    inst.QueueSave = {}
    inst.QueueLoad = {}
end
if inst.OnSave ~= nil then table.insert(inst.QueueSave, inst.OnSave) end
if inst.OnLoad ~= nil then table.insert(inst.QueueLoad, inst.OnLoad) end
inst.OnSave = OnSave
inst.OnLoad = OnLoad
end)
